/*
Copyright (c) 2010, Paul Furgale and Chi Hay Tong
All rights reserved.

Redistribution and use in source and binary forms, with or without 
modification, are permitted provided that the following conditions are 
met:

* Redistributions of source code must retain the above copyright notice, 
  this list of conditions and the following disclaimer.
* Redistributions in binary form must reproduce the above copyright 
  notice, this list of conditions and the following disclaimer in the 
  documentation and/or other materials provided with the distribution.
* The names of its contributors may not be used to endorse or promote 
  products derived from this software without specific prior written 
  permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A 
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER
OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include "fasthessian.h"
#ifndef ASRL_FASTHESSIAN_TEST_HPP
#define ASRL_FASTHESSIAN_TEST_HPP

#include "GpuSurfOctave.hpp"
#include "fasthessian.h"
#include "GpuSurfDetectorInternal.hpp"
#include <opencv/highgui.h>
#include <opencv/highgui.hpp>
#include "gpu_area.h"

/// A test suite for the timer class
BOOST_AUTO_TEST_SUITE(asrl_fasthessian)

BOOST_AUTO_TEST_CASE(test_fasthessian) 
{
  // Check that the fast hessian is deterministic.
  
  // Load an image
  cv::Mat img = cv::imread("../testdata/seq-lt-g-000001.pgm",CV_LOAD_IMAGE_GRAYSCALE);
  asrl::GpuSurfConfiguration config;

  asrl::GpuSurfDetectorInternal detector1(config);

  detector1.buildIntegralImage(img);
  detector1.detectKeypoints();

  asrl::GpuSurfDetectorInternal detector2(config);

  detector2.buildIntegralImage(img);
  detector2.detectKeypoints();

  for(unsigned i = 0 ; i < detector1.m_config.nOctaves; i++)
    {
      asrl::GpuSurfOctave & o1 = detector1.m_octaves[i];
      asrl::GpuSurfOctave & o2 = detector2.m_octaves[i];

      o1.pullHessianFromDevice();
      o2.pullHessianFromDevice();
      size_t hSize = o1.width() * o1.height() * o1.intervals();
      std::cout << "Checking the " << hSize << " elements of octave " << i << std::endl;
      BOOST_CHECK_EQUAL_COLLECTIONS( o1.h_hessian(), o1.h_hessian() + hSize, o2.h_hessian(), o2.h_hessian() + hSize );
    }

}


BOOST_AUTO_TEST_CASE(test_area_calculation)
{
  // Load an image
  cv::Mat img = cv::Mat::ones(64,64,CV_8UC1) * 126;
  asrl::GpuSurfConfiguration config;
  
  asrl::GpuSurfDetectorInternal detector1(config);

  detector1.buildIntegralImage(img);
  detector1.detectKeypoints();

  asrl::texturize_integral_image(detector1.m_intImg->d_get());    
  try {

    for(int component = 0; component < 7; component++)
      {
	asrl::eval_component( detector1.m_octaves, detector1.m_config.nOctaves, (asrl::fh_component)component);

	std::string cname = "not set";
	if(component == 0)
	  cname = "Dxx";
	else if(component == 1)
	  cname = "Dyy";
	else if(component == 2)
	  cname = "Dxx * Dyy";
	else if(component == 3)
	  cname = "Dxy";
	else if(component == 4)
	  cname = "censure";
	else if(component == 5)
	  cname = "result";
	else if(component == 6)
	  cname = "result---last bit set";
	else
	  ASRL_THROW("Unknown component: " << component);

	//for(unsigned o = 0 ; o < detector1.m_octaves.size(); o++)
	unsigned o = 0;
	  {
	    asrl::GpuSurfOctave & o1 = detector1.m_octaves[o];
	
	    o1.pullHessianFromDevice();
	    size_t hSize = o1.width() * o1.height() * o1.intervals();

	    for(int r = 1; r < o1.height() - 1; r++)
	      {
		for(int c = 1; c < o1.width() -1 ; c++)
		  {
		    for(int i = 0; i < o1.intervals(); i++)
		      {
			try
			  {
			    ASRL_ASSERT_LT(fabs(o1(r,c,i)),0.01,"FH entry " << cname << "[" << component << "]" << " at (" << o << "," << i << "," << r << "," << c << ") is too big");
			  }
			catch(std::exception const & e)
			  {
			    BOOST_ERROR(e.what());
			  }
		      }
		  }
	      }
	  }
      }
  }
  catch(std::exception const & e)
    {
      BOOST_ERROR(e.what());
    }
  asrl::fh_untexturizeIntegral();

}

BOOST_AUTO_TEST_SUITE_END()

#endif // ASRL_FASTHESSIAN_TEST_HPP
