/*
Copyright (c) 2010, Paul Furgale and Chi Hay Tong
All rights reserved.

Redistribution and use in source and binary forms, with or without 
modification, are permitted provided that the following conditions are 
met:

* Redistributions of source code must retain the above copyright notice, 
  this list of conditions and the following disclaimer.
* Redistributions in binary form must reproduce the above copyright 
  notice, this list of conditions and the following disclaimer in the 
  documentation and/or other materials provided with the distribution.
* The names of its contributors may not be used to endorse or promote 
  products derived from this software without specific prior written 
  permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A 
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER
OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include "fasthessian.h"
#ifndef ASRL_NONMAX_TEST_HPP
#define ASRL_NONMAX_TEST_HPP

#include "GpuSurfOctave.hpp"
#include "fasthessian.h"
#include "GpuSurfDetectorInternal.hpp"
#include "non_max_suppression.h"
#include <opencv/highgui.h>
#include <opencv/highgui.hpp>
#include <boost/lexical_cast.hpp>



/// A test suite for the timer class
BOOST_AUTO_TEST_SUITE(asrl_non_max_suppression)

BOOST_AUTO_TEST_CASE(test_non_max_suppression) 
{
  // Check that the fast hessian is deterministic.
  
  // Load an image
  //  cv::Mat img = cv::imread("../testdata/seq-lt-g-000001.pgm",CV_LOAD_IMAGE_GRAYSCALE);

  // Create a single octave.
  int imwidth = 101;
  int imheight = 99;
  
  const int nOctaves = 3;
  asrl::GpuSurfOctave olist[nOctaves];
  olist[0].init(imwidth, imheight, 1.f, 1.f, 1.f, 1.f, 1.f, 1.f, 0, 1, 4);
  olist[1].init(imwidth, imheight, 1.f, 1.f, 1.f, 1.f, 1.f, 1.f, 1, 1, 4);
  olist[2].init(imwidth, imheight, 1.f, 1.f, 1.f, 1.f, 1.f, 1.f, 2, 1, 4);
  init_globals(imwidth,imheight,olist,nOctaves);

  asrl::CudaSynchronizedMemory<float> hessian;
  hessian.init(olist[0].stride() * olist[0].height() * olist[0].width() * olist[0].intervals());

  asrl::GpuSurfFeatures features;
  features.featureCounterMem().pullFromDevice();
  BOOST_CHECK_EQUAL(features.featureCounterMem()[1],0);
  BOOST_CHECK_EQUAL(features.featureCounterMem()[0],0);


  float threshold = -0.1f;
 


  // Initialize the octave to zero.
  hessian.memset(0);

  dim3 threads, grid;
  threads.x = 16;
  threads.y = 4;
  threads.z = olist[0].intervals();
	
  //  grid.x = ( (olist[0].width()  + (threads.x-2) - 1) / (threads.x-2));
  grid.x = ( (olist[0].width()  + (threads.x-2) - 1) / (threads.x-2));
  grid.y = ( (olist[0].height() + (threads.y-2) - 1) / (threads.y-2));
  grid.z = 1;
  size_t sharedBytes = (threads.x+2)*threads.y*olist[0].intervals()*sizeof(float);
	
  asrl::run_surf_nonmaxonly_kernel(grid, threads, sharedBytes, hessian.d_get(), 0, features.rawFeatureMem().d_get(), features.featureCounterMem().d_get() + 1, threshold);


  features.setDirty();
  // As all entries are equal, this also tests that
  BOOST_CHECK_EQUAL(features.ftCount(),0);
  BOOST_CHECK_EQUAL(features.rawFeatureCount(),0);



  //   // Pick a couple of test cases.
  //   // This should not create a maxima as it is on a border.
  //   octave(100,100,0) = 1.f;
  //   octave(50,50,octave.intervals() - 1) = 1.f;

  try {
    for(unsigned o = 0; o <nOctaves; o++)
      {
	asrl::GpuSurfOctave & octave = olist[o];

	std::vector<cv::KeyPoint> keypoints;
	for(int i = 1; i < octave.intervals(); i++)
	  {
	    for(int r = 1; r < octave.height(); r++)
	      {
		for(int c = 1; c < octave.width(); c++)
		  {
		    features.featureCounterMem().memset(0);
		    hessian[c + olist[0].stride() * r + olist[0].stride() * octave.height() * i] = 1.f;
		    hessian.pushToDevice();
		    
		    grid.x = ( octave.width()  + (threads.x-2) - 1) / (threads.x-2);
		    grid.y = ( (octave.height() + (threads.y-2) - 1) / (threads.y-2));
		    grid.z = 1;

		    asrl::run_surf_nonmaxonly_kernel(grid, threads, sharedBytes, hessian.d_get(), o, features.rawFeatureMem().d_get(), features.featureCounterMem().d_get() + 1, threshold);
		    features.setDirty();
		    // As all entries are equal, this also tests that
		    ASRL_ASSERT_EQ(features.ftCount(),0,"at location (" << o << "," << i << "," << r << "," << c << ")");
		  
		    if(r == 0 || c == 0 || i == 0 ||
		       i == octave.intervals() -1 ||
		       r == octave.height() - 1   ||
		       c == octave.width() - 1)
		      {
			if(features.rawFeatureCount() != 0)
			  {
			    std::vector<cv::KeyPoint> keypoints;
			    features.getRawKeypoints(keypoints);
			    for(unsigned i = 0; i < keypoints.size(); i++) 
			      {
				cv::KeyPoint & k = keypoints[i];
				std::cout << "[" << i << "] (" << k.pt.x << "," << k.pt.y << ") x " << k.response << " -- " << k.octave << std::endl;
			      }
			  }

			ASRL_ASSERT_EQ(features.rawFeatureCount(),0,"at location (" << o << "," << i << "," << r << "," << c << ")");
			
		      }
		    else
		      {
			ASRL_ASSERT_EQ(features.rawFeatureCount(),1,"at location (" << o << "," << i << "," << r << "," << c << ")");
			if(features.rawFeatureCount() == 1)
			  {
			    // Check which feature it is.
			    keypoints.clear();
			    features.getRawKeypoints(keypoints);
			    ASRL_ASSERT_EQ(keypoints.size(),1,"at location (" << o << "," << i << "," << r << "," << c << ")");
			    ASRL_ASSERT_EQ(keypoints[0].pt.x,c,"at location (" << o << "," << i << "," << r << "," << c << ")");
			    ASRL_ASSERT_EQ(keypoints[0].pt.y,r,"at location (" << o << "," << i << "," << r << "," << c << ")");
			    ASRL_ASSERT_EQ(keypoints[0].octave,o,"at location (" << o << "," << i << "," << r << "," << c << ")");

			  }
		      
		      }
		    features.clearFeatureCounts();
		    hessian[c + olist[0].stride() * r + olist[0].stride() * octave.height() * i] = 0.f;

		  }
	      }
	  }
      }
  }
  catch(std::exception const & e)
    {
      BOOST_ERROR(e.what());
    }
  
  if(features.rawFeatureCount() != 0)
    {
      std::vector<cv::KeyPoint> keypoints;
      features.getRawKeypoints(keypoints);
      for(unsigned i = 0; i < keypoints.size(); i++) 
	{
	  cv::KeyPoint & k = keypoints[i];
	  std::cout << "[" << i << "] (" << k.pt.x << "," << k.pt.y << ") x " << k.response << " -- " << k.octave << std::endl;
	}
    }
  
}


  BOOST_AUTO_TEST_SUITE_END()

#endif // ASRL_NONMAX_TEST_HPP
