/*
Copyright (c) 2010, Paul Furgale and Chi Hay Tong
All rights reserved.

Redistribution and use in source and binary forms, with or without 
modification, are permitted provided that the following conditions are 
met:

* Redistributions of source code must retain the above copyright notice, 
  this list of conditions and the following disclaimer.
* Redistributions in binary form must reproduce the above copyright 
  notice, this list of conditions and the following disclaimer in the 
  documentation and/or other materials provided with the distribution.
* The names of its contributors may not be used to endorse or promote 
  products derived from this software without specific prior written 
  permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A 
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER
OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef ASRL_SURF_TESTS_HPP
#define ASRL_SURF_TESTS_HPP

#include "GpuSurfOctave.hpp"
#include "fasthessian.h"
#include "GpuSurfDetectorInternal.hpp"
#include <opencv/highgui.h>
#include <opencv/highgui.hpp>
#include <boost/lexical_cast.hpp>
#include "gpu_utils.h"

bool interesting(cv::KeyPoint const & lhs, cv::KeyPoint const & rhs)
{
  return lhs.response > rhs.response;
}

void load_keypoints(std::string const & inputKeypointFile, std::vector<cv::KeyPoint> & inKeypoints, std::vector<float> & inDescriptors, int imRows, int imCols)
{
  std::ifstream fin(inputKeypointFile.c_str());
  ASRL_ASSERT(fin.good(),"Unable to open keypoint file " << inputKeypointFile << " for reading");


  std::string line;
  std::getline(fin,line);
  int i = 1;
  while(!fin.eof())
    {
      std::istringstream lin(line);
      cv::KeyPoint k;
      lin >> k.pt.x;
      lin >> k.pt.y;
      lin >> k.size;
      lin >> k.response;
      lin >> k.angle;
      lin >> k.octave;
      int laplacian;
      lin >> laplacian;
      if(laplacian == 1)
	{
	  setLastBit(k.response);
	}
      else
	{
	  clearLastBit(k.response);
	}
      //std::cout << "read kpt: " << k.pt.x << "," << k.pt.y << ", " << k.size << "," << k.angle << std::endl;
      ASRL_ASSERT_GE_LT(k.pt.x,0,imCols,"Keypoint " << i << " is out of bounds");
      ASRL_ASSERT_GE_LT(k.pt.y,0,imRows,"Keypoint " << i << " is out of bounds");
      ASRL_ASSERT_GE_LT(k.angle,-3.15,3.15,"Keypoint " << i << " angle is out of bounds");
      ASRL_ASSERT_GE_LT(k.size,1,100,"Keypoint " << i << " size is out of bounds");
      inKeypoints.push_back(k);

      for(int d = 0; d < 64; d++)
	{
	  float di;
	  lin >> di;
	  ASRL_ASSERT(lin.good(),"Problem getting data from the string stream");
	  inDescriptors.push_back(di);
	}

      std::getline(fin,line);
      i++;
    }
}

/// A test suite for the timer class
BOOST_AUTO_TEST_SUITE(asrl_surf)

BOOST_AUTO_TEST_CASE(test_surf) 
{

  // Load an image
  cv::Mat img = cv::imread("../testdata/seq-lt-g-000001.pgm",CV_LOAD_IMAGE_GRAYSCALE);
  asrl::GpuSurfConfiguration config;

  asrl::GpuSurfDetectorInternal detector(config);

  detector.buildIntegralImage(img);
  detector.detectKeypoints();
  detector.findOrientationFast();
  detector.computeDescriptors();
  
  std::vector<cv::KeyPoint> kpGpu;
  detector.getKeypoints(kpGpu);
  // Use the octave as an index counter so we can recover the correct descriptor after the sort.
  for(int i = 0; i < kpGpu.size(); i++) kpGpu[i].octave = i;
  std::sort(kpGpu.begin(), kpGpu.end(), &interesting);

  std::vector<float> dsGpu;
  detector.getDescriptors(dsGpu);

  std::vector<cv::KeyPoint> kpFile;
  std::vector<float> dsFile;
  load_keypoints("../testdata/seq-lt-g-000001-gpusurf.key", kpFile, dsFile, img.rows, img.cols);
  // Use the octave as an index counter so we can recover the correct descriptor after the sort.
  for(int i = 0; i < kpFile.size(); i++) kpFile[i].octave = i;
  std::sort(kpFile.begin(),kpFile.end(), &interesting);

  double tolerance = 0.001;
  BOOST_REQUIRE_EQUAL(kpGpu.size(),kpFile.size());
  BOOST_REQUIRE_EQUAL(dsGpu.size(),dsFile.size());
  BOOST_REQUIRE_EQUAL(dsGpu.size(),kpGpu.size()*64);

  for(int i = 0; i < kpFile.size(); i++)
    {
      try{
	ASRL_ASSERT_EQ_TOL(kpGpu[i].pt.x,     kpFile[i].pt.x,     tolerance, "deviation from reference keypoint. i = " << i);
	ASRL_ASSERT_EQ_TOL(kpGpu[i].pt.y,     kpFile[i].pt.y,     tolerance, "deviation from reference keypoint. i = " << i);
	ASRL_ASSERT_EQ_TOL(kpGpu[i].response, kpFile[i].response, tolerance, "deviation from reference keypoint. i = " << i);
	ASRL_ASSERT_EQ_TOL(kpGpu[i].angle,    kpFile[i].angle,    tolerance, "deviation from reference keypoint. i = " << i);
	ASRL_ASSERT_EQ_TOL(kpGpu[i].size,     kpFile[i].size,     tolerance, "deviation from reference keypoint. i = " << i);
	ASRL_ASSERT_EQ(isLastBitSet(kpGpu[i].response),  isLastBitSet(kpFile[i].response), "deviation from reference keypoint. i = " << i);

	int gpuOffset = kpGpu[i].octave*64;
	int fileOffset = kpFile[i].octave*64;
	for(int d = 0; d < 64; d++)
	  {
	    ASRL_ASSERT_EQ_TOL(dsGpu[gpuOffset + d],dsFile[fileOffset + d], tolerance, "Descriptor mismatch, feature " << i << ", component " << d);
	  }
	  
      }
      catch(std::exception const & e)
	{
	  BOOST_ERROR(e.what());
	}
    }
  
  // Muck around a bit to get the descriptors.
  

}

BOOST_AUTO_TEST_SUITE_END()

#endif // ASRL_SURF_TESTS_HPP
